#include <iostream>
#include <iomanip>
#include <vector>
#include <list>
#include <algorithm>
#include <iterator>
using namespace std;

// partie Heuristique
#include "Position.hpp"
#include "Echiquier.hpp"
#include "UtilsEchiquier.cpp"
using namespace UtilsEchiquier;
#include "UtilsVector.cpp"
using namespace UtilsVector;

/** Represention du probleme */
typedef vector<int> Plateau; //! le plateau
typedef list<Plateau> Solutions; //! les solutions

/**
  Test de position valide (pas de conflit)
  @param[in] cb - un Plateau
  @param[in] k - numero de reine
  @return Vrai s'il n'y a pas de conflit de mettre la reine
  en colonne k, ligne CB[k] avec toutes les (k-1)-ieme reines
  deja placees
*/
bool issafe(const Plateau& cb, unsigned k)
{
  // La reine 0 ne peut être en prise
  // avec ses precedentes
  if (0 == k)
  {
    return true;
  }
  // Pour chaque reine deja placee
  for (unsigned j = 0; j < k; ++j)
  {
    // Delta sur les lignes
    int deltaY = cb[k] - cb[j];

    // Delta sur les colonnes: NOTE k > j donc k-j est positif
    int deltaX = static_cast<int>(k - j);

    // Si même ligne, diagonales montante ou descendante
    // alors même ligne: conflit
    if (deltaY * (deltaX - deltaY) * (deltaX + deltaY) == 0)
    {
      return false;
    }
  }
  // Ici aucun conflit
  return true;
}

/**
  Version recursive: Tente le placement de la reine
  @param[in,out] cb - un Plateau
  @param[in,out] sols - les Solutions
  @param[in] k - numero de reine
*/
void placerRN(Plateau& cb, Solutions& sols, unsigned k)
{
  // Y a-t-il n reines placees?
  // si oui, enregistre la solution
  if (k == cb.size())
  {
    sols.push_back(cb);
  }
  else
  {
    // Sinon on tente de placer la reine k
    for (unsigned j = 0; j < cb.size(); ++j)
    {
      // Place la reine colonne k, ligne j
      cb[k] = j;

      // Si pas conflit de reine avec les (k-1)-emes
      // precedentes: en placer une de plus
      if (issafe(cb, k))
      {
        placerRN(cb, sols, k+1);
      }

      // Libere la reine colonne k
      cb[k] = -1;
    }
  }
}

/**
  Lance la recherche
  @param[out] cb - un Plateau
  @param[out] sols - les Solutions
*/
void dfsRN(Plateau& cb, Solutions& sols)
{
  // Cas particulier!
  if (cb.size() == 0)
  {
    return;
  }
  // Reinitialise les donnees
  fill(cb.begin(), cb.end(), -1);
  sols.clear();

  // Lance le solveur a partir de la
  // premiere reine numero 0
  placerRN(cb, sols, 0);
}

/**
  Affiche les solutions
  @param[in] sols - les Solutions
*/
void afficherSolutions(const Solutions& sols)
{
  if (sols.empty())
  {
    cout<<"Aucune solution"<<endl;
  }
  else
  {
    cout<<"Il y a "<<sols.size()<<" solutions. On en affiche? ";
    unsigned ns;
    cin >> ns;
    Solutions::const_iterator it = sols.begin();
    for (int k = 1; k <= ns && it != sols.end(); ++k, ++it)
    {
      cout<<"Sol #"<<setw(3)<<k<<": [";
      copy(it->begin(), it->end(), ostream_iterator<int>(cout, " "));
      cout<<"]"<<endl;
    }
  }
}

/**
  @test
*/
void test_dfsRN()
{
  // Saisit la taille de l'echiquier
  cout<<"Dimension n de l'echiquier? ";
  unsigned n;
  cin >> n;
  // Definit le plateau: ligne i, colonne CB[i]
  Plateau cb(n);
  // Les solutions
  Solutions sols;
  // Lance la resolution
  dfsRN(cb, sols);
  // Affiche le resultat
  afficherSolutions(sols);
}

/**
  Version recursive: Tente le placement de la reine
  @param[in,out] cb - un Plateau
  @param[in,out] sols - les Solutions
  @param[in] k - numero de reine
*/
void placerPosRN(Plateau& cb, Solutions& sols, unsigned k)
{
  // Teste la condition d'arrêt
  if (k == cb.size())
  {
    sols.push_back(cb);
  }
  else
  {
    // Sinon lance la recherche a partir de la ligne 0
    bool placement = false;
    for (unsigned j = 0; j < cb.size() && !placement; ++j)
    {
      // Verifie si la position (j, k) n'est pas en prise;
      // si oui, invoque placer() pour poser la reine suivante
      // sur la prochaine colonne. Sinon tente le placement
      // de la reine sur la ligne suivante

      // Place la reine colonne k, ligne j
      cb[k] = j;

      // Tente le placement suivant
      if (issafe(cb, k))
      {
        placerPosRN(cb, sols, k+1);
      }

      // Libere la reine colonne k
      cb[k] = -1;
    }
  }
}

/**
  Lance la recherche, ligne pos
  @param[out] cb - un Plateau
  @param[out] sols - les Solutions
  @param[in] pos - ligne de la premiere reine
*/
void dfsPosRN(Plateau& cb, Solutions& sols, unsigned pos)
{
  // Reinitialise les donnees
  fill(cb.begin(), cb.end(), -1);
  sols.clear();

  // Place la premiere reine en (0, pos)
  cb[0] = pos;

  // Positionne les autres reines (colonnes 1 a size()-1)
  placerPosRN(cb, sols, 1);
}

/**
  "Dessine" un echiquier des reines posees
  en utilisant les symboles 'Q' ou '-'
  @param[in] cb - un Plateau
*/
void dessinerEchiquier(const Plateau& cb)
{
  // Affiche le reperage des colonnes
  cout<<"   ";
  for (unsigned x = 0; x < cb.size(); ++x)
  {
    cout<<x<<" ";
  }
  cout<<endl;

  // Affiche l'echiquier
  for (unsigned y = 0; y < cb.size(); ++y)
  {
    cout<<y<<" ";
    for (unsigned x = 0; x < cb.size(); ++x)
    {
      cout<<(static_cast<unsigned>(cb[x]) == y ? " Q" : " -");
    }
    cout<<endl;
  }
}

/**
  Dessine les solutions
  @param[in] sols - les Solutions
*/
void dessinerSolutions(const Solutions& sols)
{
  if (sols.empty())
  {
    cout<<"Aucune solution"<<endl;
  }
  else
  {
    cout<<"Il y a "<<sols.size()<<" solutions"<<endl;
    for (Solutions::const_iterator it = sols.begin(); it != sols.end(); ++it)
    {
      dessinerEchiquier(*it);
      cout<<endl;
    }
  }
}

/**
  @test
*/
void test_dfsPosRN()
{
  // Saisit la taille de l'echiquier
  cout<<"Dimension n de l'echiquier? ";
  unsigned n;
  cin >> n;
  // Definit le plateau: ligne i, colonne CB[i]
  Plateau cb(n);
  // Les solutions
  Solutions sols;
  // Demande les informations complementaires:
  cout<<"# de ligne de la reine en colonne 0? ";
  unsigned pos;
  cin >> pos;
  // Lance la resolution
  dfsPosRN(cb, sols, pos);
  // Affiche le resultat
  dessinerSolutions(sols);
}

/**
  Definit les delta du voisinage d'une case de l'echiquier
*/
Position arrDelta[] = {
  Position(-1, 0), Position(-1,-1),
  Position( 0,-1), Position( 1,-1),
  Position( 1, 0), Position( 1, 1),
  Position( 0, 1), Position(-1, 1)
};

/**
  Vecteur base sur les tableaux
*/
typedef vector<Position> CVDeplacements;
const CVDeplacements delta(&arrDelta[0], &arrDelta[8]);

/**
  Operation + sur les Position
  @param[in] pos1 - Une Position
  @param[in] pos2 - Une Position
  @return la Position pos1 + pos2
*/
Position plusPos(const Position& pos1, const Position& pos2)
{
  return Position(pos1.getX()+pos2.getX(), pos1.getY()+pos2.getY());
}

/**
  Surdefinit l'operation + sur les Position
  @param[in] pos1 - Une Position
  @param[in] pos2 - Une Position
  @return la Position pos1 + pos2
*/
Position operator+(const Position& pos1, const Position& pos2)
{
  return Position(pos1.getX()+pos2.getX(), pos1.getY()+pos2.getY());
}

/*
  Calcule l'accessibilite de chacune des positions
  @param[in] cb - un Echiquier
  @param[in,out] hb - Echiquier des accessibilites
  @param[in] delta - delta de deplacements
*/
void calculerHB(const Echiquier& cb, Echiquier& hb, const CVDeplacements& delta)
{
  for (unsigned y = 0; y < cb.sizeN(); ++y)
  {
    for (unsigned x = 0; x < cb.sizeN(); ++x)
    {
      Position pos(x, y);

      unsigned nacces = 0;
      for (unsigned nv = 0; nv < delta.size(); ++nv)
      {
        Position nvpos(pos);
        while (nvpos = nvpos + delta[nv], cb.contient(nvpos))
        {
          ++nacces;
        }
      }
      hb.set(pos, nacces);
    }
  }
}

/*
  Lance la pose des reines a partir d'une position
  Version naïve: recherche la prochaine position de pose en
  parcourant systematiquement tout l'echiquier
  @param[out] cb - un Echiquier
  @param[out] hb - Echiquier des accessibilites
  @param[out] sols - les Solutions
  @param[in] pos0 - Position de depart
*/
void placerNaif(Echiquier& cb, Echiquier& hb, vector<Position>& sols, const Position& pos0)
{
  // Reinitialise les donnees
  cb.reset(0);
  sols.clear();

  // Demarre la pose a partir de pos
  if ( not cb.contient(pos0) )
  {
    return;
  }

  // Effectue une pose tantque possible
  const unsigned NULL_ACCESS = 4 * cb.sizeN() + 1;

  Position pos = pos0;
  while(true)
  {
    // Enfile la position et la marque sur l'echiquier
    sols.push_back(pos);
    cb.set(pos, 1);

    // Marque toutes les cellules resultant de la pose en pos
    for (unsigned nv = 0; nv < delta.size(); ++nv)
    {
      // Position suivante: On part de la position courante et
      // tantque la position (suivante) est dans l'echiquier,
      // on la marque
      Position nvpos = pos;
      while (nvpos = nvpos + delta[nv], cb.contient(nvpos))
      {
        cb.set(nvpos, 1);
      }
    }

    // Recherche la prochaine position de pose de reine (position
    // non condamnee et d'accessibilite la plus petite)
    Position nextpos; // prochaine position
    unsigned nacces = NULL_ACCESS; // sa valeur d'accessibilite
    for (unsigned y = 0; y < cb.sizeN(); ++y)
    { for (unsigned x = 0; x < cb.sizeN(); ++x)
      {
        Position pos(x, y);
        if (cb.get(pos) == 0 and hb.get(pos) < nacces)
        {
          nacces = hb.get(pos) ;
          nextpos = pos;
        }
      }
    }

    // En cas d'echec: fin de la traversee
    if (NULL_ACCESS == nacces)
    {
      break;
    }
    // Sinon, on se place sur la prochaine position
    pos = nextpos;
  }
}

/**
  Version amelioree: recherche la prochaine position de pose
  en utilisant une liste de toutes les positions encore disponibles
  @param[out] cb - un Echiquier
  @param[out] hb - Echiquier des accessibilites
  @param[out] sols - les Solutions
  @param[in] pos0 - Position de depart
*/
void placerOptm(Echiquier& cb, Echiquier& hb, vector<Position>& sols, const Position& pos0)
{
  // Reinitialise les donnees
  cb.reset(0);
  sols.clear();

  // Demarre la pose a partir de pos
  if ( not cb.contient(pos0) )
  {
    return;
  }

  // Liste des positions encore disponibles.
  // au depart: toutes les positions sont non marquees
  list<Position> dispo;
  for (unsigned y = 0; y < cb.sizeN(); ++y)
  {
    for (unsigned x = 0; x < cb.sizeN(); ++x)
    {
      dispo.push_back(Position(x, y));
    }
  }

  // Effectue une pose tantque possible
  const unsigned NULL_ACCESS = 4 * cb.sizeN() + 1;

  Position pos = pos0;
  while(true)
  {
    // Enfile la position et la marque sur l'echiquier
    sols.push_back(pos);
    cb.set(pos, 1);

    // Marque toutes les cellules resultant de la pose en pos
    for (unsigned nv = 0; nv < delta.size(); ++nv)
    {
      // Position suivante: On part de la position courante et
      // tantque la position (suivante) est dans l'echiquier,
      // on la marque:
      Position nvpos = pos;
      while (nvpos = nvpos + delta[nv], cb.contient(nvpos))
      {
        cb.set(nvpos, 1);
      }
    }

    // Recherche la prochaine position de pose de reine de plus
    // petite valeur d'accessibilite: traverse la liste des
    // positions disponibles et retire celles qui ont ete
    // marquees lors de la derniere pose
    Position nextpos; // prochaine position
    unsigned nacces = NULL_ACCESS; // sa valeur d'accessibilite

    list<Position>::iterator it = dispo.begin();
    while( it != dispo.end() )
    {
      // Si la position est marquee: on la supprime des positions
      // encore disponibles, puis on passe a la position suivante.
      // Sinon: si elle est d'accessibilite plus petite, on la prend;
      // puis on passe a la position suivante
      if ( cb.get(*it) != 0)
      {
        it = dispo.erase(it);
      }
      else
      {
        unsigned a = hb.get(*it);
        if (a < nacces)
        {
          nacces = a;
          nextpos = *it;
        }
        ++it;
      }
    }

    // En cas d'echec: fin de la traversee
    if (NULL_ACCESS == nacces)
    {
      break;
    }

    // Sinon, on se place sur la prochaine position
    pos = nextpos;
  }
}

/**
  @test
*/
void test_heuristiq()
{
  // Saisit la taille de l'echiquier
  cout<<"Dimension n de l'echiquier? ";
  unsigned n;
  cin >> n;
  // Definit le plateau: ligne i, colonne CB[i]
  Echiquier cb(n);
  // Les solutions
  vector<Position> sols;
  // Tableau des accessibilites
  Echiquier hb(n);
  calculerHB(cb, hb, delta);
  // afficherEchiquier(hb, 3);
  // Lance la recherche
  for (unsigned y = 0; y < cb.sizeN(); ++y)
  {
    for (unsigned x = 0; x < cb.sizeN(); ++x)
    {
      Position pos(x, y);
      //placerNaif(cb, hb, sols, pos);
      placerOptm(cb, hb, sols, pos);

      if (sols.size() == cb.sizeN())
      {
        cout<<"Solution issue de "<<pos<<endl;
        afficherVector(sols);
      }
    }
  }
}

int main()
{
  test_dfsRN();
  test_dfsPosRN();
  test_heuristiq();
}
